"""
Generalized Simulator with Magia Module Elaboration.

This shall be used with test cases scripted with cocotb.
"""
import inspect
import os
import sys
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Sequence, Union

import cocotb.runner
from magia import Elaborator, Module


class Simulator:
    def __init__(
            self, hdl_toplevel,
            build_dir: Optional[os.PathLike] = None,
            simulator: str = "verilator",
    ):
        self.simulator_type = simulator
        self.runner = cocotb.runner.get_runner(simulator)

        self._temp_build_dir = TemporaryDirectory(prefix="sim_magia_") if build_dir is None else None
        self.build_dir = self._temp_build_dir.name if build_dir is None else build_dir

        self.output_file = f"{self.build_dir}/magia.generated.sv"
        self.top_level = hdl_toplevel
        self.source_files = []
        self.elaborated_sv = []

    def add_source(self, source_file: os.PathLike):
        """Add a SystemVerilog file for simulation."""
        self.source_files.append(source_file)

    def add_magia_module(self, *module: Module):
        """Elaborate Magia Modules for simulation."""
        self.elaborated_sv.append(
            Elaborator.to_string(*module)
        )

    def compile(self, waves: bool = False, **kwargs):
        """Compile the SV code into the simulator."""
        if waves:
            kwargs["build_args"] = kwargs.get("build_args", []) + ["--trace", "--trace-fst"]

        with open(self.output_file, mode="w") as f:
            f.write("\n\n".join(self.elaborated_sv))
        self.runner.build(
            verilog_sources=self.source_files + [self.output_file],
            hdl_toplevel=self.top_level,
            always=True,
            build_dir=self.build_dir,
            **kwargs
        )

    def sim(
            self,
            test_module: Union[str, Sequence[str]],
            python_search_path: Optional[Union[str, Sequence[str]]] = None,
            testcase: Optional[Union[str, Sequence[str]]] = None,
            waves: bool = False,
            **kwargs
    ):
        """Simulate the test module with the testcases provided."""
        if waves:
            if self.simulator_type == "verilator":
                kwargs["test_args"] = kwargs.get("test_args", []) + ["--trace", "--trace-fst"]
            else:
                kwargs["waves"] = waves

        with self._patch_sys_path(python_search_path):
            return self.runner.test(
                hdl_toplevel=self.top_level,
                test_module=test_module,
                testcase=testcase,
                build_dir=self.build_dir,
                test_dir=self.build_dir,
                results_xml="results.xml",
                **kwargs
            )

    @property
    def wave_file(self):
        """Return the path to the waveform file."""
        if self.simulator_type == "verilator":
            return Path(self.build_dir, "dump.fst")
        raise NotImplementedError("Wave file not supported for this simulator")

    @staticmethod
    def current_dir():
        """Return the directory of the caller script file."""
        caller_frame = inspect.stack()[1]
        return str(Path(caller_frame.filename).parent.absolute())

    @staticmethod
    def current_package():
        """Return the package name of the caller script file."""
        caller_frame = inspect.stack()[1]
        return str(Path(caller_frame.filename).stem)

    @staticmethod
    @contextmanager
    def _patch_sys_path(paths):
        cur_paths = sys.path
        if paths:
            if isinstance(paths, str):
                paths = [paths]
            sys.path = paths + cur_paths
        yield
        sys.path = cur_paths



"""Magia Flow CLI entry point."""
import click

import magia_flow


@click.group()
def cli():
    """Magia Flow CLI tool."""
    ...


@cli.command()
def init():
    """Create a new Magia project."""
    ...


@cli.group()
def install():
    """Install tools and binaries."""
    ...


@install.command()
@click.option("--verilator", is_flag=True, help="Install Verilator")
@click.option("--pnr", is_flag=True, help="Install PnR Tools")
def oss_cad(verilator, pnr):
    """Install OSS-CAD Suite."""
    magia_flow.oss_cad.install(verilator, pnr)


@install.command()
def surfer():
    """Install Surfer."""
    magia_flow.simulation.surfer.install()


@cli.group()
def online():
    """
    Online services.

    Elaborated SystemVerilog code will be sent to 3rd-party services.
    Don't use this command if you are developing a proprietary IP or any closed-source project.
    """
    ...


@online.group()
def local_digitaljs():
    """Commands related to local DigitalJS server."""
    ...


@local_digitaljs.group()
def start():
    """Start a local DigitalJS server."""
    magia_flow.online.digitaljs.start_local_server()


@local_digitaljs.group()
def stop():
    """Stop the local DigitalJS server."""
    magia_flow.online.digitaljs.stop_local_server()


if __name__ == "__main__":
    cli()



"""
This module provides a function to open a waveform file in the Surfer, a waveform viewer.

Project website: https://surfer-project.org/

Surfer is under the EUPL v1.2 license, which you can obtain from:
https://gitlab.com/surfer-project/surfer/-/blob/main/LICENSE-EUPL-1.2.txt?ref_type=heads
You can obtain its source code from: https://gitlab.com/surfer-project/surfer

In case you does not able to execute the Surfer check the following:
- The Surfer is available for Windows and Linux.
- `openssl` and `libssl-dev` / `openssl-dev` are required to run the Surfer on Linux.
- You may need `root` / Administrator privileges to install the Surfer, without using virtual environments.
"""

import io
import os
import platform
import shutil
import subprocess
import sys
import zipfile
from pathlib import Path

import httpx


def install():
    if shutil.which("surfer"):
        return

    system = platform.system()
    installer_url, install_dir = "", "bin"
    if system == "Windows":
        installer_url = "https://gitlab.com/api/v4/projects/42073614/jobs/artifacts/main/raw/surfer_win.zip?job=windows_build"
        install_dir = "Scripts"
    elif system == "Linux":
        installer_url = "https://gitlab.com/api/v4/projects/42073614/jobs/artifacts/main/raw/surfer_linux.zip?job=linux_build"
    else:
        raise NotImplementedError("Surfer is not available for platform: {system}")

    with httpx.stream("GET", installer_url) as response:
        byte_io = io.BytesIO()
        for chunk in response.iter_bytes():
            byte_io.write(chunk)
        byte_io.seek(0)
        with zipfile.ZipFile(byte_io) as z:
            try:
                z.extractall(Path(sys.prefix, install_dir))
            except PermissionError as err:
                raise PermissionError(
                    f"Failed to install the Surfer Binary to {str(Path(sys.prefix, install_dir))}"
                    "You may need root / Administrator privileges to install the Surfer.\n"
                    "You can also install it using a virtual environment."
                ) from err

            if system == "Linux":
                for name in z.namelist():
                    os.chmod(Path(sys.prefix, install_dir, name), 0x755)


def open_waveform(waveform):
    install()
    if not isinstance(waveform, os.PathLike):
        raise TypeError(f"waveform must be a path-like object, not {type(waveform).__name__}")

    waveform = str(waveform)
    if not waveform.endswith(".vcd") and not waveform.endswith(".fst"):
        raise ValueError(f"waveform must be a VCD or FST file, not {waveform}")
    if not os.path.exists(waveform):
        raise FileNotFoundError(f"waveform file not found: {waveform}")

    subprocess.run([shutil.which("surfer"), waveform])  # noqa: S603



"""
Using SBY to carry out Formal Verification.

SbyTask is used to declare a formal verification task.
Developer should check if SymbiYosys is installed with `sby_installed` before execution.
"""
import os
import shutil
import string
import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Iterable, Literal

import defusedxml.ElementTree as ElementTree
from magia import Elaborator


def sby_installed() -> bool:
    """Check if SymbiYosys is installed."""
    if not shutil.which("sby"):
        return False
    return True


EXTRA_FILE_TEMPLATE = string.Template("[file $fname]\n$code\n")
READ_FILE_TEMPLATE = string.Template('read -formal "$fname"\n')
SECTION_TEMPLATE = string.Template("[$section]\n$content\n")

CMD_TEMPLATE = string.Template('sby -f -d "$workdir" "$sby_file"')


class SbyTask:
    extra_code_counter = 0

    @dataclass
    class SbyTaskSpec:
        """Describes a SymbiYosys task to be executed."""

        name: str = "default"
        top_module: str = None
        mode: Literal["prove", "cover", "bmc", "live"] = "bmc"
        engines: list[str] = field(default_factory=list)
        options: list[str] = field(default_factory=list)
        script: list[str] = field(default_factory=list)
        files: list[str] = field(default_factory=list)
        extra_code: dict[str, str] = field(default_factory=dict)

        def __post_init__(self):
            if self.top_module is None:
                raise ValueError("top_module is required")
            self.script.append(f"prep -top {self.top_module}")

            if len(self.engines) == 0:
                if self.mode == "live":
                    self.engines.append("aiger suprove")
                else:
                    self.engines.append("smtbmc boolector")

            self.options = [f"mode {self.mode}"] + self.options

    @dataclass
    class SbyTaskResult:
        """Describes the result of a SymbiYosys task."""

        passed: bool
        raw_result: str = field(repr=False)
        stdout: str = field(repr=False)
        stderr: str = field(repr=False)
        tests: list[dict] = field(repr=False)
        test_count: int
        failures: int
        errors: int
        skipped: int

    def __init__(self, top_module: str, mode="bmc", work_dir=None, **kwargs):
        self.spec = self.SbyTaskSpec(top_module=top_module, mode=mode)
        self.work_dir = Path.cwd().absolute() if work_dir is None else Path(work_dir).absolute()
        self._task_executed = False
        self._result = None

    def add_file(self, *files: Iterable[os.PathLike]):
        """Add files to the task."""
        self.spec.files += [
            str(Path(file).absolute())
            for file in files
        ]

    def add_code(self, code: str, name=None):
        """Add extra code to the task."""
        if not name:
            name = f"extra_code_{self.extra_code_counter}.sv"
            self.extra_code_counter += 1

        if os.pathsep in name:
            raise ValueError("The name cannot be an absolute/relative path")
        if name in self.spec.extra_code:
            raise ValueError(f"Code with name '{name}' already exists")

        self.spec.extra_code[name] = code

    def add_script(self, *script: Iterable[str]):
        """Add script lines to the task."""
        self.spec.script += list(script)

    def _generate_script(self):
        """Generate the SymbiYosys script."""
        extra_code = [
            EXTRA_FILE_TEMPLATE.substitute(fname=fname, code=code)
            for fname, code in self.spec.extra_code.items()
        ]
        flist = self.spec.files + list(self.spec.extra_code.keys())
        read_file_scripts = [
            READ_FILE_TEMPLATE.substitute(fname=fname)
            for fname in flist
        ]

        section_formation = {
            "tasks": [f"{self.spec.name}"],
            "options": self.spec.options,
            "engines": self.spec.engines,
            "script": read_file_scripts + self.spec.script,
        }
        sections = [
                       SECTION_TEMPLATE.substitute(section=section, content="\n".join(content))
                       for section, content in section_formation.items()
                   ] + extra_code

        return "\n".join(sections)

    def run(self):
        """Run the task. Override work_dir if it is provided."""
        work_dir = self.work_dir / self.spec.top_module
        with NamedTemporaryFile("w", suffix=".sby") as sby_file:
            sby_script = self._generate_script()
            sby_file.write(sby_script)
            sby_file.flush()

            cmd = CMD_TEMPLATE.substitute(task_name=self.spec.name, workdir=str(work_dir), sby_file=sby_file.name)
            subprocess.run(cmd, shell=True, capture_output=True)  # noqa: S602

        self._task_executed = True
        self._capture_result()

    def _capture_result(self):
        """Capture the result of the task."""
        if not self._task_executed:
            raise ValueError("Run the task first")
        output_file = next(self.work_dir.rglob(f"**/*{self.spec.name}.xml"))
        raw_xml = output_file.read_text()
        root = ElementTree.fromstring(raw_xml)

        stdout = ""
        stderr = ""
        errors, failures, skipped = 0, 0, 0
        tests = []

        for testsuite in root.iter("testsuite"):
            stdout += testsuite.find("system-out").text + "\n"
            stderr += testsuite.find("system-err").text + "\n"
            for testcase in testsuite.iter("testcase"):
                new_test = {
                    "id": testcase.get("id", ""),
                    "classname": testcase.get("classname", ""),
                    "desc": testcase.get("name", ""),
                    "location": testcase.get("location", ""),
                    "status": "passed",
                    "info": {}
                }
                if testcase.find("skipped") is not None:
                    skipped += 1
                    new_test["status"] = "skipped"
                elif (node := testcase.find("failure")) is not None:
                    failures += 1
                    new_test["status"] = "failed"
                    new_test["info"] = {
                        "type": node.get("type", ""),
                        "message": node.get("message", ""),
                    }
                elif (node := testcase.find("error")) is not None:
                    errors += 1
                    new_test["status"] = "error"
                    new_test["info"] = {
                        "type": node.get("type", ""),
                        "message": node.get("message", ""),
                    }
                tests.append(new_test)

        self._result = self.SbyTaskResult(
            passed=errors == 0 and failures == 0,
            raw_result=raw_xml,
            stdout=stdout, stderr=stderr,
            test_count=len(tests),
            tests=tests,
            failures=failures, errors=errors, skipped=skipped,
        )

    @property
    def result(self):
        """Get the result of the task."""
        if not self._task_executed:
            raise ValueError("Run the task first")
        return self._result

    @classmethod
    def from_module(cls, top_module: str, *modules: Iterable[str], **kwargs):
        """Create a SbyTask from a list of modules."""
        task = cls(top_module=top_module, **kwargs)
        code = Elaborator.to_string(*modules)
        task.add_code(code, "elaborated.sv")
        return task

    @classmethod
    def from_code(cls, top_module: str, code: str, **kwargs):
        """Create a SbyTask from a code string."""
        task = cls(top_module=top_module, **kwargs)
        task.add_code(code, "manual_code.sv")
        return task



"""
Integrating Magia to other online services.

Code under this package is responsible for integrating Magia to other online services.

WARNING:
Elaborated SystemVerilog code will be sent to 3rd-party services.
The service party will have access to the code and may store it for their own purposes.
Don't use this package if you are developing a proprietary IP or any closed-source project.
"""

from .digitaljs import elaborate_on_digitaljs

__all__ = [
    "elaborate_on_digitaljs"
]



"""Utility functions related to Verilator."""
import os
import shutil
import subprocess
import sys
from pathlib import Path


def init_verilator():
    """Initialize Verilator in virtual environment for cocotb tests."""
    if not (verilator := shutil.which("verilator")):
        raise RuntimeError("Verilator not found in PATH")

    version = subprocess.check_output([verilator, "--version"]).decode().strip().split()[1]  # noqa: S603
    if float(version) < 5.006:
        raise RuntimeError("Verilator version >= 5.006 is required")

    # Skip if not in venv
    if sys.base_prefix == sys.prefix:
        return

    # Check if verilator is already copied into venv
    verilator_mounted = Path(sys.prefix, "bin", "verilator").exists()
    if not verilator_mounted:
        # Locate Verilator binaries and the original VERILATOR_ROOT
        verilator_bin_location = Path(shutil.which("verilator_bin")).parent
        verilator_root = Path(
            subprocess.check_output([verilator, "--getenv", "VERILATOR_ROOT"]).decode().strip()  # noqa: S603
        )
        verilator_bins = list(verilator_bin_location.glob("verilator*"))

        # Copy files into venv
        shutil.copytree(verilator_root, Path(sys.prefix), dirs_exist_ok=True)
        for path in verilator_bins:
            shutil.copy(path, Path(sys.prefix, "bin"))

        # Patch verilated.mk
        # Replace the Python3 interpreter with the one in venv
        makefile_in = Path(sys.prefix, "include", "verilated.mk")
        makefile_content = makefile_in.read_text().splitlines()
        for i, line in enumerate(makefile_content):
            if line.startswith("PYTHON3 ="):
                makefile_content[i] = f"PYTHON3 = {sys.executable}"
                break
        makefile_in.write_text("\n".join(makefile_content))

    # Override VERILATOR_ROOT
    os.environ["VERILATOR_ROOT"] = sys.prefix



import subprocess
import time
import webbrowser
from pathlib import Path

import httpx
from magia import Elaborator


def elaborate_on_digitaljs(*modules, server="https://digitaljs.tilk.eu"):
    """
    Elaborate the given modules and open the result in the DigitalJS web app.

    Cautions:
    The elaborated SystemVerilog code will be sent to the DigitalJS service.
    It will be stored on the server and may be used for their own purposes.

    Visit https://github.com/tilk/digitaljs_online and deploy a local instance
    if you are concerned about privacy and want to keep the elaboration local.
    """
    sv_code = Elaborator.to_string(*modules)

    synthesis_payload = {
        "files": {
            "top.sv": sv_code
        },
        "options": {"optimize": True, "fsm": "yes", "fsmexpand": True, "lint": False},
    }

    limits = httpx.Limits(max_keepalive_connections=1, max_connections=1)
    client = httpx.Client(limits=limits)

    synthesis_rsp = client.post(f"{server}/api/yosys2digitaljs", json=synthesis_payload)
    synthesis_rsp = synthesis_rsp.json()
    if "error" in synthesis_rsp:
        failure_msg = "\n".join([synthesis_rsp["error"], synthesis_rsp.get("yosys_stderr", "")])
        raise RuntimeError(f"Synthesis failed: \n {failure_msg}")

    result = synthesis_rsp["output"]
    time.sleep(1)

    store_rsp = client.post(f"{server}/api/storeCircuit", json=result)
    if store_rsp.is_error:
        raise RuntimeError(f"Failed to store circuit: {store_rsp.text}")
    store_rsp = store_rsp.json()
    webbrowser.open_new(f"{server}/#{store_rsp}")


def start_local_server():
    """
    Start a local DigitalJS server.

    This function requires docker compose installed.
    """
    work_dir = (Path(__file__).parent / "digitaljs").absolute()
    docker_compose = _docker_compose_cmd()
    subprocess.run(docker_compose + ["up", "-d", "--build"], cwd=work_dir)  # noqa: S603, S607
    webbrowser.open_new("http://localhost:3000")


def stop_local_server():
    """
    Start a local DigitalJS server.

    This function requires docker compose installed.
    """
    work_dir = (Path(__file__).parent / "digitaljs").absolute()
    docker_compose = _docker_compose_cmd()
    subprocess.run(docker_compose + ["down"], cwd=work_dir)  # noqa: S603, S607


def _docker_compose_cmd():
    try:
        res = subprocess.run(["docker", "compose", "version"], capture_output=True, text=True)  # noqa: S603, S607
        if res.returncode == 0:
            return ["docker", "compose"]
    except FileNotFoundError:
        pass

    try:
        res = subprocess.run(["docker-compose", "--version"], capture_output=True, text=True)  # noqa: S603, S607
        if res.returncode == 0:
            return ["docker-compose"]
    except FileNotFoundError:
        pass
    raise RuntimeError("docker compose is not installed")



import logging
import platform
import sys
import tarfile
from tempfile import NamedTemporaryFile

import httpx


def install(
        verilator: bool = False,
        pnr: bool = False,
):
    logger = logging.getLogger(__name__)

    system = platform.system()

    if system != "Linux":
        raise NotImplementedError("Only Linux is supported for now.")
    if sys.prefix == sys.base_prefix:
        raise RuntimeError("You must use a virtual environment to install the OSS-CAD Suite.")

    installer_url = (
        "https://github.com/YosysHQ/oss-cad-suite-build/releases/download/"
        "2024-02-28/oss-cad-suite-linux-x64-20240228.tgz"
    )

    logger.info("Downloading the OSS-CAD Binary...")
    with httpx.stream("GET", installer_url, follow_redirects=True) as response, NamedTemporaryFile(
            suffix=".tgz") as temp_file:
        for chunk in response.iter_bytes(chunk_size=16 * 1024 * 1024):
            temp_file.write(chunk)
        temp_file.flush()

        logger.info("Extracting files...")
        with tarfile.open(temp_file.name, mode="r:gz") as t:
            members = [
                tinfo.replace(name=tinfo.name.removeprefix("oss-cad-suite/"))
                for tinfo in t.getmembers()
                if (tname := tinfo.name) != "oss-cad-suite" and
                   "__pycache__" not in tname and
                   ("verilator" not in tname or verilator) and
                   "examples" not in tname and (
                           (
                                   "nextpnr" not in tname and
                                   "icebox" not in tname and
                                   "prjoxide" not in tname and
                                   "trellis" not in tname
                           ) or pnr
                   )
            ]
            t.extractall(sys.prefix, members)  # noqa: S202
